#!/usr/bin/env python

from .techman_client import *
from .stateless_client import *
from .stateful_client import *
from .tmsta_client import *
from .tmsvr_client import *
from .tmsct_client import *



#!/usr/bin/env python

from .stateless_packet import *
from ..exceptions import *

class CPERR_code:

   SUCCESS=0x00
   PACKET_ERR=0x01
   CHECKSUM_ERR=0x02
   HEADER_ERR=0x03
   DATA_ERR=0x04
   NO_LISTEN=0xF1

   @staticmethod
   def description(value):
      if value == CPERR_code.SUCCESS: return 'Success'
      if value == CPERR_code.PACKET_ERR: return 'Packet error'
      if value == CPERR_code.CHECKSUM_ERR: return 'Checksum error'
      if value == CPERR_code.HEADER_ERR: return 'Header error'
      if value == CPERR_code.DATA_ERR: return 'Packet data error'
      if value == CPERR_code.NO_LISTEN: return 'Listen Node is not active'
      return 'Unknown error'

class CPERR_packet(StatelessPacket):

   HEADER='CPERR'

   def __init__(self, *args):
      try:
         # Instantiated with StatelessPacket object
         if isinstance(args[0], StatelessPacket):
            self._header = args[0]._header
            self._data = args[0]._data
         # Instantiated with raw packet data
         elif not isinstance(args[0], int): super().__init__(*args)
         # Instantiated with payload data
         else:
            self._header = self.HEADER
            self._data = self._encode_data(args[0])
      except: raise TMParseError() from None

   def _encode_data(self, error_code):
      return f'{error_code:02X}'

   def _decode_data(self, data):
      return int(data, 16)

   @property
   def value(self): return self._decode_data(self._data)

   @property
   def description(self): return CPERR_code.description(self.value)



#!/usr/bin/env python

import asyncio

from .stateless_client import *
from ..packets import *
from ..exceptions import *

class TMSTA_client(StatelessClient):

   PORT=5890

   def __init__(self, *, robot_ip, conn_timeout=3):
      super().__init__(robot_ip=robot_ip, robot_port=self.PORT, conn_timeout=conn_timeout)

   def _on_connection(self, reader, writer):
      return TMSTA_connection(reader, writer, self._conn_timeout)

class TMSTA_connection(StatelessConnection):

   async def is_listen_node_active(self):
      # Build TMSTA packet
      req = TMSTA_packet(TMSTA_type.IN_LISTEN_MODE, None)
      # Submit
      res = TMSTA_packet(await self.send(req))
      # Parse response
      assert res.ptype == TMSTA_type.IN_LISTEN_MODE
      return res.params[0]

   async def get_queue_tag_status(self, tag_id):
      # Build TMSTA packet
      req = TMSTA_packet(TMSTA_type.QUEUE_TAG, [tag_id])
      # Submit
      res = TMSTA_packet(await self.send(req))
      # Parse response
      assert res.ptype == TMSTA_type.QUEUE_TAG
      return res.params[1]



#!/usr/bin/env python

import ast

from .stateful_packet import *
from ..exceptions import *

class TMSVR_type:

   RESPONSE_STATUS=0
   VALUE_DATA=2
   VALUE_REQUEST=12

class TMSVR_status:

   SUCCESS=(0x00, 'OK')
   FORMAT_ERROR=(0x01, 'NotSupport')
   IP_NO_WRITE=(0x02, 'WritePermission')
   FORMAT_MISMATCH=(0x03, 'InvalidData')
   ITEM_NOEXIST=(0x04, 'NotExist')
   WRITE_PROTECTED=(0x05, 'ReadOnly')
   MODE_PROTECTED=(0x06, 'ModeError')
   VARTYPE_MISMATCH=(0x07, 'ValueError')

   @staticmethod
   def value(value):
      if value == TMSVR_status.SUCCESS[0]: return TMSVR_status.SUCCESS
      if value == TMSVR_status.FORMAT_ERROR[0]: return TMSVR_status.FORMAT_ERROR
      if value == TMSVR_status.IP_NO_WRITE[0]: return TMSVR_status.IP_NO_WRITE
      if value == TMSVR_status.FORMAT_MISMATCH[0]: return TMSVR_status.FORMAT_MISMATCH
      if value == TMSVR_status.ITEM_NOEXIST[0]: return TMSVR_status.ITEM_NOEXIST
      if value == TMSVR_status.WRITE_PROTECTED[0]: return TMSVR_status.WRITE_PROTECTED
      if value == TMSVR_status.MODE_PROTECTED[0]: return TMSVR_status.MODE_PROTECTED
      if value == TMSVR_status.VARTYPE_MISMATCH[0]: return TMSVR_status.VARTYPE_MISMATCH

   @staticmethod
   def description(status):
      if status == TMSVR_status.SUCCESS: return 'Success'
      if status == TMSVR_status.FORMAT_ERROR: return 'The communication format or mode is not supported'
      if status == TMSVR_status.IP_NO_WRITE: return 'The connected client is not permitted to write'
      if status == TMSVR_status.FORMAT_MISMATCH: return 'The communication format and the data content format are mismatched'
      if status == TMSVR_status.ITEM_NOEXIST: return 'Item to write or read does not exist'
      if status == TMSVR_status.WRITE_PROTECTED: return 'Unable to write to read-only items'
      if status == TMSVR_status.MODE_PROTECTED: return 'Can\'t write in the current control mode, try switching to auto mode'
      if status == TMSVR_status.VARTYPE_MISMATCH: return 'Values to write mismatches with the configured type or the size.'

class TMSVR_packet(StatefulPacket):

   HEADER='TMSVR'

   def __init__(self, *args):
      try:
         if len(args) == 1:
            # Instantiated with StatefulPacket object
            if isinstance(args[0], StatefulPacket):
               self._header = args[0]._header
               self._data = args[0]._data
            # Instantiated with raw packet data
            else: super().__init__(*args)
         # Instantiated with payload data
         else:
            self._header = self.HEADER
            self._data = self._encode_data(*args)
      except: raise TMParseError() from None

   def _encode_data(self, *args):
      encoded = super()._encode_data(args[0])
      ptype = args[1]
      encoded += f'{ptype},'
      if ptype == TMSVR_type.VALUE_REQUEST:
         for variable in args[2]:
            encoded += f'{variable}\r\n'
         encoded = encoded[:len(encoded)-2]
      elif ptype == TMSVR_type.VALUE_DATA:
         for variable, value in args[2].items():
            encoded += f'{variable}={self._encode_value(value)}\r\n'
         encoded = encoded[:len(encoded)-2]
      elif ptype == TMSVR_type.RESPONSE_STATUS:
         encoded += f'{args[2][0]:02X},'
         if len(args) == 3: encoded += args[2][1]
         else: encoded += f'{args[2][1]};{args[3]}'
      return encoded

   def _decode_data(self, data):
      ptype = int(data[self._find_nth(data, ',', 1)+1:self._find_nth(data, ',', 2)])
      payload = data[self._find_nth(data, ',', 2)+1:]
      # Correct protocol bug
      if ptype == TMSVR_type.VALUE_REQUEST and '=' in payload: ptype = TMSVR_type.VALUE_DATA
      # Decode payload based on packet type
      if ptype == TMSVR_type.VALUE_REQUEST:
         return ptype, set(payload.split('\r\n'))
      if ptype == TMSVR_type.VALUE_DATA:
         items = {}
         for varval in payload.split('\r\n'):
            var, val = varval.split('=')
            items[var] = self._decode_value(val)
         return ptype, items
      if ptype == TMSVR_type.RESPONSE_STATUS:
         status = TMSVR_status.value(int(payload[:payload.find(',')], 16))
         errdata = None if ';' not in payload else payload[payload.find(';')+1:]
         return ptype, status, errdata

   def _encode_value(self, value):
      if isinstance(value, str): return f'"{value}"'
      if isinstance(value, list):
         return str(value).replace(' ', '').replace('[', '{').replace(']', '}')
      if isinstance(value, bool) and value == True: return 'true'
      if isinstance(value, bool) and value == False: return 'false'
      return str(value)

   def _decode_value(self, value):
      if value == 'true': return True
      if value == 'false': return False
      if '"' in value: return value[1:len(value)-1]
      if '{' in value: return ast.literal_eval(value.replace('{', '[').replace('}', ']'))
      return ast.literal_eval(value)

   @property
   def ptype(self): return self._decode_data(self._data)[0]

   @property
   def status(self): return self._decode_data(self._data)[1]

   @property
   def errdata(self): return self._decode_data(self._data)[2]

   @property
   def errdesc(self): return TMSVR_status.description(self.status)

   @property
   def items(self): return self._decode_data(self._data)[1]



#!/usr/bin/env python

import asyncio

from .stateful_client import *
from ..packets import *
from ..exceptions import *

class TMSVR_client(StatefulClient):

   PORT=5891

   def __init__(self, *, robot_ip, client_id='SVRpy', conn_timeout=3):
      super().__init__(robot_ip=robot_ip, robot_port=self.PORT, client_id=client_id, conn_timeout=conn_timeout)
      self._client_id = client_id

   def _on_connection(self, reader, writer):
      return TMSVR_connection(self._client_id, reader, writer, self._conn_timeout)

class TMSVR_connection(StatefulConnection):

   def add_broadcast_callback(self, broadcast_callback):
      def parsed_callback(packet): broadcast_callback(TMSVR_packet(packet).items)
      super().add_broadcast_callback(parsed_callback)

   async def _execute(self, packet):
      # Submit
      res = TMSVR_packet(await self.send(packet))
      # Parse response
      assert res.handle_id == packet.handle_id
      if res.ptype == TMSVR_type.RESPONSE_STATUS and res.status != TMSVR_status.SUCCESS:
         if res.errdata is None: raise TMSVRError(res.errdesc)
         else: raise TMSVRError(f'{res.errdesc} (\'{res.errdata}\')')
      return res.items

   async def get_values(self, items):
      req = TMSVR_packet(self._obtain_handle_id(), TMSVR_type.VALUE_REQUEST, items)
      return await self._execute(req)

   async def get_value(self, key):
      try: return (await self.get_values({key}))[key]
      except KeyError: raise TMSVRError(f'Response did not contain value \'{key}\'')

   async def set_values(self, items):
      req = TMSVR_packet(self._obtain_handle_id(), TMSVR_type.VALUE_DATA, items)
      return await self._execute(req)

   async def set_value(self, key, value):
      return await self.set_values({key: value})



#!/usr/bin/env python

import ast

from .stateful_packet import *
from ..exceptions import *

class TMSCT_type:

   REQUEST=0
   RESPONSE=1

class TMSCT_status:

   SUCCESS='OK'
   ERROR='ERROR'

class TMSCT_command_type:

   FUNCTION=0
   VARIABLE=1

class TMSCT_packet(StatefulPacket):

   HEADER='TMSCT'

   def __init__(self, *args):
      try:
         if len(args) == 1:
            # Instantiated with StatefulPacket object
            if isinstance(args[0], StatefulPacket):
               self._header = args[0]._header
               self._data = args[0]._data
            # Instantiated with raw packet data
            else: super().__init__(*args)
         # Instantiated with payload data
         else:
            self._header = self.HEADER
            self._data = self._encode_data(*args)
      except: raise TMParseError() from None

   def _encode_data(self, *args):
      encoded = super()._encode_data(args[0])
      ptype = args[1]
      if ptype == TMSCT_type.REQUEST:
         commands = '\r\n'.join(list(map(self._encode_command, args[2])))
         return encoded + commands
      if ptype == TMSCT_type.RESPONSE:
         if len(args[3]) == 0: return encoded + args[2]
         return encoded + f'{args[2]};{";".join(list(map(str, args[3])))}'

   def _decode_data(self, data):
      payload = data[data.find(',')+1:]
      status = TMSCT_status.SUCCESS if TMSCT_status.SUCCESS in payload else None
      if status is None: status = TMSCT_status.ERROR if TMSCT_status.ERROR in payload else None
      ptype = TMSCT_type.REQUEST if status is None else TMSCT_type.RESPONSE
      if ptype == TMSCT_type.REQUEST:
         commands = payload.split('\r\n')
         return ptype, list(map(self._decode_command, commands))
      else:
         parts = payload.split(';')
         lines = [] if len(parts) == 1 else list(map(int, parts[1:]))
         return ptype, status, lines

   def _encode_command(self, command):
      if len(command) == 2 and command[0] == TMSCT_command_type.VARIABLE: return command[1]
      name = command[0] if len(command) == 2 else command[1]
      arglist = command[1] if len(command) == 2 else command[2]
      argstr = str(arglist).replace(' ', '').replace('[', '{').replace(']', '}')
      argstr = argstr.replace('True', 'true').replace('False', 'false').replace('None', 'none')
      argstr = argstr.replace('\'', '"')
      return f'{name}({argstr[1:-1]})'

   def _decode_command(self, command):
      if '(' not in command: return TMSCT_command_type.VARIABLE, command
      name = command[:command.find('(')]
      args = command[command.find('(')+1:command.find(')')].replace(' ', '')
      args = args.replace('{', '[').replace('}', ']')
      args = args.replace('true', 'True').replace('false', 'False').replace('none', 'None')
      return TMSCT_command_type.FUNCTION, name, ast.literal_eval('[' + args + ']')

   @property
   def ptype(self): return self._decode_data(self._data)[0]

   @property
   def commands(self): return self._decode_data(self._data)[1]

   @property
   def status(self): return self._decode_data(self._data)[1]

   @property
   def lines(self): return self._decode_data(self._data)[2]



#!/usr/bin/env python

import sys

from ..exceptions import *

class StatelessPacket:

   def __init__(self, *args):
      try:
         self._header, self._data = None, None
         if len(args) == 1:
            self._header, self._data = self._decode(args[0])
         elif len(args) == 2:
            self._header = args[0]
            self._data = args[1]
      except: raise TMParseError() from None

   def _encode(self, header, data):
      checksum_data = f'{header},{len(data)},{data},'
      checksum = 0b0
      utf_str = checksum_data.encode('utf-8')
      for byte in utf_str: checksum ^= byte
      return f'${checksum_data}*{checksum:02X}\r\n'

   def _decode(self, packet):
      if isinstance(packet, bytes): packet = packet.decode('utf-8')
      header = packet[1:self._find_nth(packet, ',', 1)]
      data_length = int(packet[self._find_nth(packet, ',', 1)+1:self._find_nth(packet, ',', 2)])
      data = packet[self._find_nth(packet, ',', 2)+1:self._find_nth(packet, ',', 2)+1+data_length]
      return header, data

   def _find_nth(self, msg, pattern, n):
      start = msg.find(pattern)
      while start >= 0 and n > 1:
         start = msg.find(pattern, start+len(pattern))
         n -= 1
      return start

   def encoded(self): return self._encode(self._header, self._data).encode('utf-8')



#!/usr/bin/env python

import asyncio
from asyncio import CancelledError

from .techman_client import *
from ..packets import *
from ..exceptions import *

class StatefulClient(TechmanClient):

   def __init__(self, *, robot_ip, robot_port, client_id='SCpy', conn_timeout=3):
      super().__init__(robot_ip=robot_ip, robot_port=robot_port, conn_timeout=conn_timeout)
      self._client_id = client_id

   def _on_connection(self, reader, writer):
      return StatefulConnection(self._client_id, reader, writer, self._conn_timeout)

class StatefulConnection(TechmanConnection):

   def __init__(self, client_id, reader, writer, conn_timeout):
      super().__init__(reader, writer, conn_timeout)
      self._client_id = client_id

   def _init_vars(self):
      super()._init_vars()
      self._listen_task = None
      self._requests = []
      self._in_listen = False
      self._listen_is_awaited = False
      self._broadcast_callback = None
      self._msg_cnt = 0
      self._quit_cond = None

   def add_broadcast_callback(self, broadcast_callback):
      if not self._in_listen: self._start_listen()
      self._broadcast_callback = broadcast_callback

   async def keep_alive(self, quit=None):
      self._listen_is_awaited = True
      self._quit_cond = quit
      await self._listen_task

   def quit(self):
      def quit(): return True
      self._quit_cond = quit

   async def send(self, packet):
      # Send message
      req_fut = asyncio.Future()
      self._requests.append([packet, req_fut, 0])
      if not self._in_listen: self._start_listen()
      self._writer.write(packet.encoded())
      await self._writer.drain()
      # Wait until result
      return await req_fut

   def _obtain_handle_id(self):
      handle_id = f'{self._client_id}{self._msg_cnt}'
      self._msg_cnt += 1
      return handle_id

   def _handle_exception(self, exc):
      for request in self._requests: request[1].set_exception(exc)
      self._requests = []
      if self._listen_is_awaited: raise exc from None

   def _on_message(self, packet):
      # Check if this is a response to ongoing request
      handle_id = packet.handle_id
      index, req_fut = -1, None
      for i, request in enumerate(self._requests):
         if request[0].handle_id != handle_id: continue
         index, req_fut = i, request[1]      
      # Call request callback
      if index != -1:
         del self._requests[index]
         req_fut.set_result(packet)
      # Call broadcast callback
      elif self._broadcast_callback is not None: self._broadcast_callback(packet)

   def _start_listen(self):
      if self._in_listen: return
      self._in_listen = True
      def callback(res):
         # For debugging purposes
         try: res.result()
         except CancelledError as e: pass
         except Exception as e: 
            if not self._listen_is_awaited: print(f'Exception caught: {type(e)}')
      self._listen_task = asyncio.ensure_future(self._listen())
      self._listen_task.add_done_callback(callback)

   async def _listen(self):
      try:
         while True:
            read_bytes = None
            while read_bytes is None:
               if self._quit_cond is not None and self._quit_cond(): break
               try: read_bytes = await asyncio.wait_for(self._reader.read(100000), timeout=self._conn_timeout)
               except asyncio.TimeoutError: pass
            if self._quit_cond is not None and self._quit_cond(): break
            # Empty byte indicates lost connection
            if read_bytes == b'': raise TMConnectError(None, msg='Socket connection was closed by server')
            # Skip corrupt packet
            elif len(read_bytes) > 5 and (read_bytes[0] != 0x24 or read_bytes[-5] != 0x2A): continue
            # Valid respone
            res = StatefulPacket(read_bytes)
            if res._header == 'CPERR':
               self._handle_exception(TMProtocolError(CPERR_packet(res).description))
               continue
            self._on_message(res)
            # Refresh request if no answer within 5 packets
            for request in self._requests:               
               if request[2] > 5:
                  request[2] = 0
                  self._writer.write(request[0].encoded())
                  await self._writer.drain()
               else: request[2] += 1
            # Quit loop if we are done
            if self._broadcast_callback is None and len(self._requests) == 0: break
            if self._quit_cond is not None and self._quit_cond(): break
         self._in_listen = False
      except CancelledError as e: raise e # Delegate asyncio exception
      except TechmanException as e: self._handle_exception(e)
      except ConnectionError as e: self._handle_exception(TMConnectError(e))
      except Exception as e: self._handle_exception(e)



#!/usr/bin/env python

from .exceptions import *
from .packets import *
from .clients import *

def connect_sta(*, robot_ip, conn_timeout=3):
    return TMSTA_client(robot_ip=robot_ip, conn_timeout=conn_timeout)

def connect_svr(*, robot_ip, client_id='SVRpy', conn_timeout=3):
    return TMSVR_client(robot_ip=robot_ip, client_id=client_id, conn_timeout=conn_timeout)

def connect_sct(*, robot_ip, client_id='SCTpy', conn_timeout=3, suppress_warns=False):
    return TMSCT_client(robot_ip=robot_ip, client_id=client_id, conn_timeout=conn_timeout, suppress_warns=suppress_warns)



#!/usr/bin/env python

import asyncio
from asyncio import CancelledError

from .techman_client import *
from ..packets import *
from ..exceptions import *

class StatelessClient(TechmanClient):

   def _on_connection(self, reader, writer):
      return StatelessConnection(reader, writer, self._conn_timeout)

class StatelessConnection(TechmanConnection):

   async def send(self, packet):
      try:
         # Send packet
         self._writer.write(packet.encoded())
         await self._writer.drain()
         # Wait for response
         read_bytes = await asyncio.wait_for(self._reader.read(100000), timeout=self._conn_timeout)
         # Empty byte indicates lost connection
         if read_bytes == b'': raise TMConnectError(None, msg='Socket connection was closed by server')
         # Validate response
         res = StatelessPacket(read_bytes)
         if res._header == 'CPERR': raise TMProtocolError(CPERR_packet(res).description)
         else: return res
      except CancelledError as e: raise e # Delegate asyncio exception
      except TechmanException as e: raise e
      except asyncio.TimeoutError: raise TMConnectError(None, msg='Did not receive a message from server') from None
      except ConnectionError as e: raise TMConnectError(e) from None
      except RuntimeError as e:
         if 'coroutine' in str(e): raise TMConnectError(None, msg='Only one stateless outgoing request can be active at any time!') from None



#!/usr/bin/env python

import sys
import asyncio

from .stateful_client import *
from ..packets import *
from ..exceptions import *

class TMSCT_commands:

   # ==== general commmands ====

   def exit_listen(self):
      return ('ScriptExit', [])

   def set_queue_tag(self, tag_id, wait_for_completion=False):
      return ('QueueTag', [tag_id, 1 if wait_for_completion else 0])

   def wait_for_queue_tag(self, tag_id, timeout=-1):
      return ('WaitQueueTag', [tag_id, int(timeout)])

   def stop_motion(self):
      return ('StopAndClearBuffer', [])

   def pause_project(self):
      return ('Pause', [])

   def resume_project(self):
      return ('Resume', [])

   def set_base(self, base):
      if isinstance(base, list) and len(base) != 6:
         raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('ChangeBase', [base])

   def set_tcp(self, tcp, weight=None, inertia=None):
      if isinstance(tcp, list) and len(tcp) != 6:
         raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      arglist = [tcp]
      if weight is not None: arglist.append(weight)
      if inertia is not None: arglist.append(inertia)
      return ('ChangeTCP', arglist)

   def set_load_weight(self, weight):
      return ('ChangeLoad', [weight])

   # ==== PVT ====

   def enter_point_pvt_mode(self):
      return ('PVTEnter', [1])

   def add_pvt_point(self, tcp_point_goal, tcp_point_velocities_goal, duration):
      if len(tcp_point_goal) != 6 or len(tcp_point_velocities_goal) != 6:
         raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('PVTPoint', [tcp_point_goal, tcp_point_velocities_goal, duration/1000.0])

   def enter_joint_pvt_mode(self):
      return ('PVTEnter', [0])

   def add_pvt_joint_angles(self, joint_angles_goal, joint_angle_velocities_goal, duration):
      if len(joint_angles_goal) != 6 or len(joint_angle_velocities_goal) != 6:
         raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('PVTPoint', [joint_angles_goal, joint_angle_velocities_goal, duration/1000.0])

   def exit_pvt_mode(self):
      return ('PVTExit', [])

   def pause_pvt_mode(self):
      return ('PVTPause', [])

   def resume_pvt_mode(self):
      return ('PVTResume', [])

   # ==== PTP ====

   def move_to_point_ptp(self, tcp_point_goal, speed_perc, acceleration_duration, blending_perc=0.0, use_precise_positioning=False, pose_goal=None):
      if speed_perc > 1.0 or blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(tcp_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      arglist = ['CPP', tcp_point_goal, int(100 * speed_perc), int(acceleration_duration), int(100 * blending_perc), not use_precise_positioning]
      if pose_goal is not None: arglist.append(pose_goal)
      return ('PTP', arglist)

   def move_to_relative_point_ptp(self, relative_point_goal, speed_perc, acceleration_duration, blending_perc=0.0, relative_to_tcp=False, use_precise_positioning=False):
      if speed_perc > 1.0 or blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(relative_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('Move_PTP', ['TPP' if relative_to_tcp else 'CPP', relative_point_goal, int(100 * speed_perc), int(acceleration_duration), int(100 * blending_perc), not use_precise_positioning])

   def move_to_joint_angles_ptp(self, joint_angles_goal, speed_perc, acceleration_duration, blending_perc=0.0, use_precise_positioning=False):
      if speed_perc > 1.0 or blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(joint_angles_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('PTP', ['JPP', joint_angles_goal, int(100 * speed_perc), int(acceleration_duration), int(100 * blending_perc), not use_precise_positioning])

   def move_to_relative_joint_angles_ptp(self, relative_joint_angles_goal, speed_perc, acceleration_duration, blending_perc=0.0, use_precise_positioning=False):
      if speed_perc > 1.0 or blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(relative_joint_angles_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('Move_PTP', ['JPP', relative_joint_angles_goal, int(100 * speed_perc), int(acceleration_duration), int(100 * blending_perc), not use_precise_positioning])

   # ==== Path ====

   def move_to_point_path(self, tcp_point_goal, velocity, acceleration_duration, blending_perc=0.0):
      if blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(tcp_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('PLine', ['CAP', tcp_point_goal, int(velocity), int(acceleration_duration), int(100 * blending_perc)])

   def move_to_relative_point_path(self, relative_point_goal, velocity, acceleration_duration, blending_perc=0.0, relative_to_tcp=False):
      if blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(relative_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('Move_PLine', ['TAP' if relative_to_tcp else 'CAP', relative_point_goal, int(velocity), int(acceleration_duration), int(100 * blending_perc)])

   def move_to_joint_angles_path(self, joint_angles_goal, velocity, acceleration_duration, blending_perc=0.0):
      if blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(joint_angles_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('PLine', ['JAP', joint_angles_goal, int(velocity), int(acceleration_duration), int(100 * blending_perc)])

   def move_to_relative_joint_angles_path(self, relative_joint_angles_goal, velocity, acceleration_duration, blending_perc=0.0):
      if blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(relative_joint_angles_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      return ('Move_PLine', ['JAP', relative_joint_angles_goal, int(velocity), int(acceleration_duration), int(100 * blending_perc)])

   # ==== Line ====

   def move_to_point_line(self, tcp_point_goal, speed, acceleration_duration, blending=0.0, speed_is_velocity=False, blending_is_radius=False, use_precise_positioning=False):
      if (not speed_is_velocity and speed > 1.0) or (not blending_is_radius and blending > 1.0): raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(tcp_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      data_mode = f'C{"A" if speed_is_velocity else "P"}{"R" if blending_is_radius else "P"}'
      speed = int(speed if speed_is_velocity else (100 * speed))
      blending = int(blending if blending_is_radius else (100 * blending))
      return ('Line', [data_mode, tcp_point_goal, speed, int(acceleration_duration), blending, not use_precise_positioning])

   def move_to_relative_point_line(self, relative_point_goal, speed, acceleration_duration, blending=0.0, relative_to_tcp=False, speed_is_velocity=False, blending_is_radius=False, use_precise_positioning=False):
      if (not speed_is_velocity and speed > 1.0) or (not blending_is_radius and blending > 1.0): raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(relative_point_goal) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      data_mode = f'{"T" if relative_to_tcp else "C"}{"A" if speed_is_velocity else "P"}{"R" if blending_is_radius else "P"}'
      speed = int(speed if speed_is_velocity else (100 * speed))
      blending = int(blending if blending_is_radius else (100 * blending))
      return ('Move_Line', [data_mode, relative_point_goal, speed, int(acceleration_duration), blending, not use_precise_positioning])

   # ==== Circle ====

   def move_on_circle(self, tcp_point_1, tcp_point_2, speed, acceleration_duration, arc_angle=0, blending_perc=0.0, speed_is_velocity=False, use_precise_positioning=False):
      if (not speed_is_velocity and speed > 1.0) or blending_perc > 1.0: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): percentages should have value between 0.0 and 1.0')
      if len(tcp_point_1) != 6 or len(tcp_point_2) != 6: raise TMSCTError(f'{sys._getframe().f_code.co_name}(): position array should have exactly 6 elements')
      speed = int(speed if speed_is_velocity else (100 * speed))
      return ('Circle', ['CAP' if speed_is_velocity else 'CPP', tcp_point_1, tcp_point_2, speed, int(acceleration_duration), int(100 * blending_perc), arc_angle, not use_precise_positioning])

class TMSCT_client(StatefulClient):

   PORT=5890

   def __init__(self, *, robot_ip, client_id='SCTpy', conn_timeout=3, suppress_warns=False):
      super().__init__(robot_ip=robot_ip, robot_port=self.PORT, client_id=client_id, conn_timeout=conn_timeout)
      self._client_id = client_id
      self._suppress_warns = suppress_warns

   def _on_connection(self, reader, writer):
      return TMSCT_connection(self._client_id, reader, writer, self._conn_timeout, self._suppress_warns)

class TMSCT_connection(StatefulConnection, TMSCT_commands):

   def __init__(self, client_id, reader, writer, conn_timeout, suppress_warns):
      super().__init__(client_id, reader, writer, conn_timeout)
      self._suppress_warns = suppress_warns

   def start_transaction(self): return TMSCT_transaction(self)

   async def _execute(self, commands):
      if not isinstance(commands, list): commands = [commands]
      commands = list(map(self._unfold_command, commands))
      # Build TMSCT packet
      handle_id = self._obtain_handle_id()
      req = TMSCT_packet(handle_id, TMSCT_type.REQUEST, commands)
      # Submit
      res = TMSCT_packet(await self.send(req))
      # Parse response
      assert res.handle_id == handle_id
      assert res.ptype == TMSCT_type.RESPONSE
      if len(res.lines) > 0:
         commands = [req.commands[i-1] for i in res.lines]
         enc_cmnds = list(map(req._encode_command, commands))
         if res.status == TMSCT_status.SUCCESS:
            if not self._suppress_warns:
               if len(res.lines) == 1:
                  print(f'[TMSCT_client] WARN: The command \'{enc_cmnds[0]}\' resulted in a warning')
               elif len(res.lines) > 1: print(f'[TMSCT_client] WARN: The following commands resulted in a warning: {enc_cmnds}')
         if res.status == TMSCT_status.ERROR:
            if len(res.lines) == 1: raise TMSCTError(f'The command \'{enc_cmnds[0]}\' resulted in an error')
            else: raise TMSCTError(f'The following commands resulted in an error: {enc_cmnds}')

   def _flatten(self, seq, scalarp=None):
      for item in seq:
         if item is None or isinstance(item, str) or not hasattr(item, '__iter__'): yield item
         else: yield from self._flatten(item)

   def _unfold_command(self, command):
      return (command[0], list(self._flatten(command[1])))

   def __getattribute__(self, attr):
      if hasattr(TMSCT_commands, attr):
         def caller(*args, **kwargs): return self._meta_execute(getattr(TMSCT_commands, attr), *args, **kwargs)
         return caller
      else: return super().__getattribute__(attr)

   async def _meta_execute(self, command_func, *args, **kwargs):
      return await self._execute(command_func(self, *args, **kwargs))

class TMSCT_transaction(TMSCT_commands):

   def __init__(self, conn):
      self._conn = conn
      self._transaction = []
      self._did_submit = False

   async def submit(self):
      if self._did_submit: raise TMSCTError('Transaction was already submitted')
      self._did_submit = True
      return await self._conn._execute(self._transaction)

   def __getattribute__(self, attr):
      if hasattr(TMSCT_commands, attr):
         def caller(*args, **kwargs): return self._meta_execute(getattr(TMSCT_commands, attr), *args, **kwargs)
         return caller
      else: return super().__getattribute__(attr)

   def _meta_execute(self, command_func, *args, **kwargs):
      if self._did_submit: raise TMSCTError('Transaction was already submitted')
      self._transaction.append(command_func(self, *args, **kwargs))



#!/usr/bin/env python

from .stateless_packet import *
from ..exceptions import *

class TMSTA_type:

   IN_LISTEN_MODE=0
   QUEUE_TAG=1

class TMSTA_packet(StatelessPacket):

   HEADER='TMSTA'

   def __init__(self, *args):
      try:
         if len(args) == 1:
            # Instantiated with StatelessPacket object
            if isinstance(args[0], StatelessPacket):
               self._header = args[0]._header
               self._data = args[0]._data
            # Instantiated with raw packet data
            else: super().__init__(*args)
         # Instantiated with payload data
         elif len(args) == 2:
            self._header = self.HEADER
            self._data = self._encode_data(args[0], args[1])
      except: raise TMParseError() from None

   def _encode_data(self, ptype, params):
      if isinstance(ptype, int): ptype = self._encode_int(ptype)
      if params is None: return f'{ptype}'
      params = ','.join(list(map(self._encode_param, params)))
      return f'{ptype},{params}'

   def _decode_data(self, data):
      if ',' not in data: return self._decode_int(data), None
      ptype = self._decode_int(data[:data.find(',')])
      params = data[data.find(',')+1:].split(',')
      return ptype, list(map(self._decode_param, params))

   def _encode_param(self, param):
      if param is None: return 'none'
      if isinstance(param, bool) and param == True: return 'true'
      if isinstance(param, bool) and param == False: return 'false'
      if isinstance(param, int): return self._encode_int(param)
      return param

   def _decode_param(self, param):
      if param == 'none': return None
      if param == 'true': return True
      if param == 'false': return False
      if len(param) == 0: return param
      if param[0] == '0' or param[0] == '1': return self._decode_int(param)
      return param

   def _encode_int(self, value):
      if value >= 10: return str(value)
      else: return '0' + str(value)

   def _decode_int(self, str):
      if str[0] == '0': return int(str[1])
      else: return int(str)

   @property
   def ptype(self): return self._decode_data(self._data)[0]

   @property
   def params(self): return self._decode_data(self._data)[1]



#!/usr/bin/env python

import sys

from .stateless_packet import *
from ..exceptions import *

class StatefulPacket(StatelessPacket):

   def __init__(self, *args):
      try: super().__init__(*args)
      except: raise TMParseError() from None

   def _encode_data(self, handle_id):
      return f'{handle_id},'

   @property
   def handle_id(self):
      return self._data[:self._data.find(',')]



import setuptools
import sys

if sys.version_info[:2] < (3, 8): raise Exception("'techmanpy' requires Python >= 3.8")

with open('README.md', 'r', encoding='utf-8') as fh:
   long_description = fh.read()

setuptools.setup(
   name='techmanpy',
   version='1.1',
   description='Communication driver for Techman robots',
   long_description=long_description,
   long_description_content_type='text/markdown',
   author='Jules van der Toorn',
   author_email='julesvandertoorn@gmail.com',
   license='MIT',
   classifiers=[
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.8',
      'License :: OSI Approved :: MIT License',
      'Operating System :: POSIX :: Linux',
      'Intended Audience :: Science/Research',
      'Natural Language :: English'
   ],
   keywords='techman driver manipulator robotics',
   url='https://github.com/jvdtoorn/techmanpy',
   project_urls={
      'Source': 'https://github.com/jvdtoorn/techmanpy'
   },
   packages=setuptools.find_packages(),
   package_dir={'': '.'},
   python_requires='~=3.8',
)


#!/usr/bin/env python

import asyncio

from ..exceptions import *

class TechmanClient:

   def __init__(self, *, robot_ip, robot_port, conn_timeout=3):
      self._robot_ip = robot_ip
      self._robot_port = robot_port
      self._conn_timeout = conn_timeout
      self._connection = None

   def _on_connection(self, reader, writer):
      return TechmanConnection(reader, writer, self._conn_timeout)
   
   async def __aenter__(self):
      try:
         reader, writer = await asyncio.wait_for(asyncio.open_connection(self._robot_ip, self._robot_port), timeout=self._conn_timeout)
         self._connection = self._on_connection(reader, writer)
         return self._connection
      except ConnectionError as e: raise TMConnectError(e) from None
      except asyncio.TimeoutError: raise TMConnectError(None, msg=f'Opening connection timed out after {self._conn_timeout} seconds') from None

   async def __aexit__(self, exc_type, exc_msg, _): self._connection._close()

class TechmanConnection:

   def __init__(self, reader, writer, conn_timeout):
      self._conn_timeout = conn_timeout
      self._reader = reader
      self._writer = writer
      self._init_vars()
   
   def _init_vars(self): pass

   def _close(self): self._writer.close()



#!/usr/bin/env python


class TechmanException(Exception):
   ''' Base class for exceptions. '''
   
   def __init__(self):
      super().__init__()
      self._msg = 'TechmanException: Something went wrong'

   def __str__(self): return str(self._msg).strip()

class TMConnectError(TechmanException):
   ''' Raised at connection related errors '''
   
   def __init__(self, exc, msg=None):
      super().__init__()
      if msg is not None: self._msg = f'TMConnectError: {msg}'
      else: self._msg = f'TMConnectError: {exc} ({type(exc).__name__})'

class TMParseError(TechmanException):
   ''' Raised at packet parse related errors '''

   def __init__(self):
      super().__init__()
      self._msg = 'TMParseError: Could not parse packet'

class TMProtocolError(TechmanException):
   ''' Raised at protocol related errors '''
   
   def __init__(self, msg):
      super().__init__()
      self._msg = f'TMProtocolError: {msg}'

class TMSTAError(TMProtocolError):
   ''' Raised at STA protocol related errors '''
   
   def __init__(self, msg):
      super().__init__(msg)
      self._msg = f'TMSTAError: {msg}'

class TMSVRError(TMProtocolError):
   ''' Raised at SVR protocol related errors '''
   
   def __init__(self, msg):
      super().__init__(msg)
      self._msg = f'TMSVRError: {msg}'

class TMSCTError(TMProtocolError):
   ''' Raised at SCT protocol related errors '''
   
   def __init__(self, msg):
      super().__init__(msg)
      self._msg = f'TMSCTError: {msg}'
